# *****************************************************************************
# * Copyright by ams OSRAM AG                                                 *
# * All rights are reserved.                                                  *
# *                                                                           *
# * IMPORTANT - PLEASE READ CAREFULLY BEFORE COPYING, INSTALLING OR USING     *
# * THE SOFTWARE.                                                             *
# *                                                                           *
# * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       *
# * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT         *
# * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS         *
# * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
# * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,     *
# * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT          *
# * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES LOSS OF USE,      *
# * DATA, OR PROFITS OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY      *
# * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT       *
# * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE     *
# * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.      *
# *****************************************************************************

"""Create systemd service for TMF8806 ZeroMQ server."""

from pathlib import Path
from zipfile import ZipFile


def create_systemd_service(service_name, python_script_path, log_file_path):
    """Create and start a systemd service for a Python script.

    Args:
        service_name (str): The name of the systemd service.
        python_script_path (str): The path to the Python script to run as a service.
        log_file_path (str): The path to the file where the output of the service will be logged.

    Returns:
        None

    Side effects:
        Creates a systemd service file at /etc/systemd/system/{service_name}.service
        Enables and starts the systemd service using systemctl
        Prints a message indicating the status of the service creation and start
    """
    service_file_path = f"/etc/systemd/system/{service_name}.service"
    with open(service_file_path, "w", encoding="utf-8") as f:
        f.write("[Unit]\n"
                f"Description={service_name}\n"
                "After=multi-user.target\n"
                "\n"
                "[Service]\n"
                "Type=simple\n"
                "Restart=always\n"
                f"ExecStart=/usr/bin/python3 {python_script_path}\n"
                f"StandardOutput=truncate:{log_file_path}\n"
                f"StandardError=truncate:{log_file_path}\n"
                "\n"
                "[Install]\n"
                "WantedBy=multi-user.target\n"
                )

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Install th TMF8806 ZeroMQ server as systemd service.")
    parser.add_argument("pkg_path", help="Path to the TMF8806 server zip file.")
    parser.add_argument("--dest", help="Install destination path", default="/home/ams/zmq_server")
    parser.add_argument("--name", help="Name of the service.", default="tmf8806_zmq_server")
    parser.add_argument("--log", help="Path of the service log file.", default="/var/log/tmf8806_zmq_server.log")
    args = parser.parse_args()

    with ZipFile(args.pkg_path) as f:
        f.extractall(args.dest)

    create_systemd_service(args.name, Path(args.dest, "tmf8806_zmq_server.py"), args.log)
