# *****************************************************************************
# * Copyright by ams OSRAM AG                                                 *
# * All rights are reserved.                                                  *
# *                                                                           *
# * IMPORTANT - PLEASE READ CAREFULLY BEFORE COPYING, INSTALLING OR USING     *
# * THE SOFTWARE.                                                             *
# *                                                                           *
# * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       *
# * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT         *
# * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS         *
# * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
# * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,     *
# * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT          *
# * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES LOSS OF USE,      *
# * DATA, OR PROFITS OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY      *
# * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT       *
# * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE     *
# * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.      *
# *****************************************************************************
"""
Linux driver Adapter

Adapter classes to connect ZeroMQ server with Raspberry PI EVM.
"""

import ctypes
import logging
import re
import shutil
import struct
import subprocess
from enum import IntEnum
from io import BytesIO
from pathlib import Path
from queue import Empty, Queue
from threading import Thread
from time import sleep
from typing import BinaryIO, Iterable, List, Optional, Tuple, Union
from zipfile import ZipFile

from .tmf8x0x_zeromq_common import (
    EVM_VERSION,
    BaseAdapterInterface,
    CommandError,
    DeviceInfo,
    EvmCtrlId,
    HistogramConfig,
    HistogramFlags,
    HistogramsAndResult,
    PersistenceThresholdConfig,
)
from .tmf8806_regs import (
    struct__tmf8806MeasureCmdAlgoSettings,
    struct__tmf8806MeasureCmdDataSettings,
    struct__tmf8806MeasureCmdGpioSettings,
    struct__tmf8806MeasureCmdSnrVcselSpreadSpecSettings,
    struct__tmf8806MeasureCmdSpreadSpectrumSpadChargePump,
    struct__tmf8806MeasureCmdSpreadSpectrumVcselChargePump,
    tmf8806DistanceResultFrame,
    tmf8806FactoryCalibData,
    tmf8806MeasureCmd,
)

__all__ = [
    "LinuxAdapter"
]

logger = logging.getLogger(__name__)


PIN_HEADER_DRIVER_PATH = "/sys/class/i2c-adapter/i2c-1/1-0041/"
CSI_PORT_DRIVER_PATH = "/sys/class/i2c-adapter/i2c-0/0-0041/"

UPDATE_FOlDER = "/tmp/tmf8806/update"
DRIVER_MODULE = "/opt/USBSensorBridgeRuntime/modules/tmf8806.ko"


class FrameID(IntEnum):
    """Output frame IDs"""
    HIST_ELECTRICAL_CAL = 0x01
    HIST_PROXIMITY = 0x04
    HIST_DISTANCE = 0x07
    HIST_ALG_PILEUP = 0x10
    HIST_ALG_PU_TDC_SUM = 0x11
    RESULT = 0x55


class LinuxAdapter(BaseAdapterInterface):
    """
    Adapter to Linux driver.

    Args:
        port: Optional, root path to the Linux driver.
              If `None` then the default driver path for pin header or
              CSI port is used.
    Raise:
        FileNotFoundError: Driver path not found.
    """

    def __init__(self, drv_path: Optional[str] = None) -> None:
        if drv_path is None:
            if Path(CSI_PORT_DRIVER_PATH).exists():
                self._drv_path = Path(CSI_PORT_DRIVER_PATH)
            elif Path(PIN_HEADER_DRIVER_PATH).exists():
                self._drv_path = Path(PIN_HEADER_DRIVER_PATH)
            else:
                raise FileNotFoundError("TMF8806 driver path not found.")
        else:
            self._drv_path = Path(drv_path)
            if not self._drv_path.exists():
                raise FileNotFoundError(f"{self._drv_path} doesn't exist.")

        self._current_hist_config = HistogramConfig()
        self._device_info = DeviceInfo()

        self._distance_result_byte_size = ctypes.sizeof(tmf8806DistanceResultFrame)
        self._frame_header_struct = struct.Struct("<2BH")
        self._single_hist = struct.Struct("<256H")
        self._summed_hist = struct.Struct("<128H")
        self._abort = False
        self._read_thread: Optional[Thread] = None
        self._results = Queue()

    def _set_configuration(self, config: tmf8806MeasureCmd) -> None:
        """
        Write configuration data to driver.

        Args:
            config: Measurement configuration.
        """
        logger.debug("Write measurement configuration.")
        with open(self._drv_path / "tmf8806_app0/spreadSpecSpad", mode="w", encoding="utf-8") as f:
            f.write(f"0x{bytes(config.data.spreadSpecSpadChp)[0]:x}")

        with open(self._drv_path / "tmf8806_app0/spreadSpecVcsel", mode="w", encoding="utf-8") as f:
            f.write(f"0x{bytes(config.data.spreadSpecVcselChp)[0]:x}")

        with open(self._drv_path / "tmf8806_app0/data_setting", mode="w", encoding="utf-8") as f:
            f.write(f"0x{bytes(config.data.data)[0]:x}")

        with open(self._drv_path / "tmf8806_app0/alg_setting", mode="w", encoding="utf-8") as f:
            f.write(f"0x{bytes(config.data.algo)[0]:x}")

        with open(self._drv_path / "tmf8806_app0/gpio_setting", mode="w", encoding="utf-8") as f:
            f.write(f"0x{bytes(config.data.gpio)[0]:x}")

        with open(self._drv_path / "tmf8806_app0/capture_delay", mode="w", encoding="utf-8") as f:
            f.write(f"{config.data.daxDelay100us}")

        with open(self._drv_path / "tmf8806_app0/snr", mode="w", encoding="utf-8") as f:
            f.write(f"0x{bytes(config.data.snr)[0]:x}")

        with open(self._drv_path / "tmf8806_app0/period", mode="w", encoding="utf-8") as f:
            f.write(f"{config.data.repetitionPeriodMs}")

        with open(self._drv_path / "tmf8806_app0/iterations", mode="w", encoding="utf-8") as f:
            f.write(f"{config.data.kIters}")

    def _read_result_buffer(self, fp: BinaryIO, size: int) -> Union[bytes, None]:
        """
        Read the requested number of bytes from output data FIFO.

        Args:
            fp: File pointer to the result output file.

            size: Bytes to read.

        Returns:
            Requested number of bytes or None if the measurement is stopped.
        """
        buffer = fp.read(size)
        while len(buffer) < size and not self._abort:
            sleep(0.01)
            buffer += fp.read(size - len(buffer))

        if self._abort:
            buffer = None

        return buffer

    def _read_result_frame(self, fp: BinaryIO) -> Union[None, Tuple[FrameID, int, bytes]]:
        """
        Read one result frame from the file.

        Args:
            fp: File pointer to the result output file.

        Returns:
            Frame identifier, Frame number, Frame payload or `None` if the measurement is stopped.
        """
        buffer = self._read_result_buffer(fp, self._frame_header_struct.size)
        if buffer is None:
            return None

        # Read frame header
        frame_id, frame_number, payload_size = self._frame_header_struct.unpack(buffer)

        if payload_size > 0:
            payload = self._read_result_buffer(fp, payload_size)
        else:
            payload = b""

        if payload is None:
            return None
        return FrameID(frame_id), frame_number, payload

    def _decode_and_scale_hist(self, buffer: bytes, hist_type: HistogramFlags) -> Union[List[int], List[Iterable[int]]]:
        """
        Decode and scale histogram data from byte buffer.

        Args:
            buffer: Histogram data as byte buffer.

            hist_type: Type of histogram data.

        Returns:
            Scaled histogram data.
        """
        if hist_type == HistogramFlags.ALG_PU_TDC_SUM:
            scaled_histogram = [e * 4 for e in self._summed_hist.unpack(buffer)] + [0] * 128

        elif hist_type == HistogramFlags.ALG_PILEUP:
            scaled_histogram = [e * 2 for e in self._summed_hist.unpack(buffer)] + [0] * 128

        else:
            scaled_histogram = [
                [e << hist[127] for e in hist[0:127]] + list(hist[127:128])
                + [e << hist[255] for e in hist[128:255]] + list(hist[255:256])
                for hist in self._single_hist.iter_unpack(buffer)
            ]

        return scaled_histogram

    def _read_results(self) -> None:
        """
        Read result data from driver and process them.
        """
        with open(self._drv_path / "tmf8806_app0/app0_tof_output", mode="rb") as f:
            # Clean residual data
            _ = f.read()
            result = HistogramsAndResult()
            expected_frame_number = None
            while not self._abort:
                frame = self._read_result_frame(f)
                if frame is None:
                    break
                # Read output frame
                frame_id, frame_number, payload = frame

                # Check frame number
                if expected_frame_number is None:
                    expected_frame_number = frame_number
                elif expected_frame_number >= 256:
                    expected_frame_number = 0

                if frame_number == expected_frame_number:
                    expected_frame_number += 1
                else:
                    logger.warning("Lost output data frame. Expect frame number %i received %i",
                                   expected_frame_number, frame_number)
                    expected_frame_number = frame_number + 1
                    logger.info("Reset expected frame number to %i", expected_frame_number)

                # EC histogram
                if FrameID.HIST_ELECTRICAL_CAL == frame_id:
                    result.histogramsEc = self._decode_and_scale_hist(payload, HistogramFlags.ELECTRICAL_CAL)

                # Prox histogram
                elif FrameID.HIST_PROXIMITY == frame_id:
                    result.histogramsProx = self._decode_and_scale_hist(payload, HistogramFlags.PROXIMITY)

                # Distance histogram
                elif FrameID.HIST_DISTANCE == frame_id:
                    result.histogramsDist = self._decode_and_scale_hist(payload, HistogramFlags.DISTANCE)

                # ALG_PILEUP histogram
                elif FrameID.HIST_ALG_PILEUP == frame_id:
                    result.histogramsDistPuc.append(self._decode_and_scale_hist(payload, HistogramFlags.ALG_PILEUP))

                # ALG_PU_TDC_SUM histogram
                elif FrameID.HIST_ALG_PU_TDC_SUM == frame_id:
                    result.histogramSum = self._decode_and_scale_hist(payload, HistogramFlags.ALG_PU_TDC_SUM)

                # Distance result
                elif FrameID.RESULT == frame_id:
                    result.result = tmf8806DistanceResultFrame.from_buffer_copy(payload[4:])  # Remove 4 status bytes
                    self._results.put(result)
                    result = HistogramsAndResult()
                else:
                    logger.error("Invalid output frame ID received: %i", frame_id)

    def _read_device_info(self) -> None:
        """
        Read the device information.

        Note:
            Measurement must be stopped before reading
            the device information.

        Raises:
            CommandError: Reading device information failed
        """
        self._device_info.evm_controller_id = EvmCtrlId.RASPI
        self._device_info.evm_version = EVM_VERSION

        with open(self._drv_path / "tmf8806_common/program_version", encoding="utf-8") as f:
            version_data = f.read()
        if not version_data:
            raise CommandError("Version data are empty.")
        fw_version = [int(e, 0) for e in version_data.strip().split()]
        self._device_info.fw_version = int.from_bytes(bytes(int(e) for e in fw_version[0:4]), "little")

        driver_version = self._read_driver_version()
        self._device_info.host_driver_version = int.from_bytes(
            driver_version[0].to_bytes(1, "little")
            + driver_version[1].to_bytes(1, "little")
            + b'\x00\x00',
            'little')

        self._device_info.sensor_serial = self._read_sensor_serial()

    def open(self) -> None:
        """
        Open the device connection.

        Raises:
            CommandError: Open device connection failed.
        """
        self.stop_measurement()
        self._read_device_info()

    def close(self) -> None:
        """
        Close the device connection.

        Raises:
            CommandError: Closing device connection failed.
        """
        self.stop_measurement()

    def identify(self) -> DeviceInfo:
        """
        Read the device information.

        Returns:
            Device information.
        """
        return self._device_info

    def _read_driver_version(self) -> Tuple[int, int]:
        """
        Read the sensor Linux driver version.

        Returns:
            Major and minor version of the Linux driver.
        """
        driver_info = subprocess.run(["modinfo", DRIVER_MODULE], check=True, capture_output=True).stdout.decode()
        version = re.search(r"version:\s*(\d+)\.(\d+)", driver_info)
        if version is not None:
            major = int(version.group(1))
            minor = int(version.group(2))
        else:
            major = 0
            minor = 0
        return major, minor

    def _read_sensor_serial(self) -> int:
        """
        Read the sensor serial number.

        Returns:
            Sensor serial number
        """
        serial = 0
        # Send read serial command
        with open(self._drv_path / "tmf8806_common/register_write", mode="w", encoding="utf-8") as f:
            f.write("0x10:0x47")
        sleep(0.1)
        # Read serial
        with open(self._drv_path / "tmf8806_common/registers", mode="r", encoding="utf-8") as f:
            for line in f:
                if line.startswith("0x28:"):
                    serial_bytes = line.strip().replace("0x28: ", "", 1).replace("0x", "").split(" ")
                    serial_bytes.reverse()
                    serial = int("".join(e.rjust(2, "0") for e in serial_bytes), 16)
        return serial

    def start_measurement(self,
                          config: tmf8806MeasureCmd,
                          cal_data: Optional[tmf8806FactoryCalibData] = None) -> None:
        """
        Start measurement.

        Args:
            config: Measurement configuration.

            cal_data: Optional, calibration data.

        Raise:
            CommandError: Starting measurement failed.
        """
        self._set_configuration(config=config)
        if cal_data is not None:
            self._set_calibration_data(cal_data)
        self._abort = False
        if self._read_thread is None:
            self._read_thread = Thread(target=self._read_results, name="Result data reader")
            self._read_thread.start()
        with open(self._drv_path / "tmf8806_app0/capture", mode="w", encoding="utf-8") as f:
            f.write("1")

    def stop_measurement(self) -> None:
        """
        Stop measurement.

        Raise:
            CommandError: Stopping measurement failed.
        """
        self._abort = True
        if self._read_thread is not None:
            self._read_thread.join(5.0)
            if self._read_thread.is_alive():
                logger.warning("Result read thread is still alive.")
            self._read_thread = None
        with open(self._drv_path / "tmf8806_app0/capture", mode="w", encoding="utf-8") as f:
            f.write("0")

    def get_configuration(self) -> tmf8806MeasureCmd:
        """
        Get measurement configuration.

        Return:
            Measurement configuration.
        """
        cfg = tmf8806MeasureCmd()

        with open(self._drv_path / "tmf8806_app0/spreadSpecSpad", encoding="utf-8") as f:
            value = int(f.read(), base=16)
            cfg.data.spreadSpecSpadChp = struct__tmf8806MeasureCmdSpreadSpectrumSpadChargePump.from_buffer_copy(
                value.to_bytes(1, 'little'))

        with open(self._drv_path / "tmf8806_app0/spreadSpecVcsel", encoding="utf-8") as f:
            value = int(f.read(), base=16)
            cfg.data.spreadSpecVcselChp = struct__tmf8806MeasureCmdSpreadSpectrumVcselChargePump.from_buffer_copy(
                value.to_bytes(1, 'little'))

        with open(self._drv_path / "tmf8806_app0/data_setting", encoding="utf-8") as f:
            value = int(f.read(), base=16)
            cfg.data.data = struct__tmf8806MeasureCmdDataSettings.from_buffer_copy(value.to_bytes(1, 'little'))

        with open(self._drv_path / "tmf8806_app0/alg_setting", encoding="utf-8") as f:
            value = int(f.read(), base=16)
            cfg.data.algo = struct__tmf8806MeasureCmdAlgoSettings.from_buffer_copy(value.to_bytes(1, 'little'))

        with open(self._drv_path / "tmf8806_app0/gpio_setting", encoding="utf-8") as f:
            value = int(f.read(), base=16)
            cfg.data.gpio = struct__tmf8806MeasureCmdGpioSettings.from_buffer_copy(value.to_bytes(1, 'little'))

        with open(self._drv_path / "tmf8806_app0/capture_delay", encoding="utf-8") as f:
            cfg.data.daxDelay100us = int(f.read(), base=10)

        with open(self._drv_path / "tmf8806_app0/snr", encoding="utf-8") as f:
            value = int(f.read(), base=16)
            cfg.data.snr = struct__tmf8806MeasureCmdSnrVcselSpreadSpecSettings.from_buffer_copy(
                value.to_bytes(1, 'little'))

        with open(self._drv_path / "tmf8806_app0/period", encoding="utf-8") as f:
            cfg.data.repetitionPeriodMs = int(f.read(), base=10)

        with open(self._drv_path / "tmf8806_app0/iterations", encoding="utf-8") as f:
            cfg.data.kIters = int(f.read(), base=10)

        return cfg

    def get_data(self, timeout: float = 1.0) -> Union[HistogramsAndResult, None]:
        """
        Read result data from device.

        Args:
            timeout: Maximal wait time for new result data. If no new result data available
                `None`shall be returned.

        Returns:
            Result data.
        """
        try:
            result = self._results.get(block=True, timeout=timeout)
        except Empty:
            result = None

        return result

    def get_histogram_config(self) -> HistogramConfig:
        """
        Get histogram dumping configuration.

        Returns:
            Histogram dumping configuration.
        """
        return HistogramConfig(buffer=self._current_hist_config.to_buffer())

    def set_histogram_config(self, config: HistogramConfig) -> None:
        """
        Set histogram configuration.

        Args:
            config: Histogram configuration.
        """
        logger.debug("Set histogram cfg %s", config)
        self._current_hist_config.from_buffer(config.to_buffer())
        with open(self._drv_path / "tmf8806_app0/app0_histogram_readout", mode="w", encoding="utf-8") as f:
            value = f"0x{config.to_flags():x}"
            logger.debug("Set histogram config: %s", value)
            f.write(value)

    def get_persistence_threshold_config(self) -> PersistenceThresholdConfig:
        """
        Get persistence and threshold configuration.

        Returns:
            Persistence and threshold configuration.
        """
        with open(self._drv_path / "tmf8806_app0/distance_thresholds", encoding="utf-8") as f:
            values = f.read().strip().split()
        return PersistenceThresholdConfig(persistence=int(values[0], base=10),
                                          low_threshold=int(values[1], base=10),
                                          high_threshold=int(values[2], base=10),
                                          )

    def set_persistence_threshold_config(self, config: PersistenceThresholdConfig) -> None:
        """
        Set persistence and threshold configuration.

        Args:
            config: Persistence and threshold configuration.
        """
        with open(self._drv_path / "tmf8806_app0/distance_thresholds", mode="w", encoding="utf-8") as f:
            f.write(f"{config.persistence} {config.low_threshold} {config.high_threshold}")

    def run_factory_calibration(self, config: tmf8806MeasureCmd) -> tmf8806FactoryCalibData:
        """
        Run factory calibration.

        Args:
            config: Configuration used for factory calibration.

        Returns:
            Calibration data.

        Raise:
            CommandError: Factory calibration failed.
        """
        self._set_configuration(config=config)

        logger.debug("Start factory calibration")
        with open(self._drv_path / "tmf8806_app0/app0_apply_fac_calib", mode="w", encoding="utf-8") as f:
            f.write("1")

        sleep(1.5)  # Wait for calibration data

        # Read result
        logger.debug("Read factory calibration result.")
        calib_data = self._get_calibration_data()
        return calib_data

    def _get_calibration_data(self) -> tmf8806FactoryCalibData:
        """
        Read calibration data.

        Returns:
            Calibration data.
        """
        with open(self._drv_path / "tmf8806_app0/app0_fac_calib", encoding="utf-8") as f:
            values = f.read()
        logger.debug("Get calibration data: %s", values)
        buffer = b"".join(int(e, 16).to_bytes(1, "little") for e in values.strip().split())
        calib_data = tmf8806FactoryCalibData.from_buffer_copy(buffer)
        return calib_data

    def _set_calibration_data(self, data: tmf8806FactoryCalibData) -> None:
        """
        Write calibration data.

        Args:
            data: Calibration data.
        """
        value = " ".join(hex(e) for e in bytes(data))
        logger.debug("Set calibration data: %s", value)
        with open(self._drv_path / "tmf8806_app0/app0_fac_calib", mode="w", encoding="utf-8") as f:
            f.write(value)

    def update_target_binaries(self, zip_blob: bytes) -> None:
        """
        Update the EVM software like:

            - Sensor firmware
            - Linux driver
            - ZeroMQ server

        The needed files are zip as one file. The zip file must include one file
        with the name `update.py` in the root. This file must contain all instructions to
        update the related software components and restart the system if needed.
        On server side the zip file will be unzipped. And the `update.py` will be
        executed in a new process.

        Args:
            zip_blob: Zip file as byte blob.
        """
        logger.info("Update target binaries")
        update_folder = Path(UPDATE_FOlDER)
        if update_folder.exists():
            shutil.rmtree(update_folder) # cleanup
        update_folder.mkdir(parents=True, exist_ok=True)
        logger.debug("Created and cleaned update folder %s", update_folder)
        # Unzip
        with ZipFile(BytesIO(zip_blob)) as f:
            f.extractall(UPDATE_FOlDER)
        logger.debug("Files unzipped.")

        # Start update
        cmds = ["python3", "update.py"]
        subprocess.Popen(cmds, start_new_session=True, cwd=update_folder)
        logger.debug("Update process started.")
