# *****************************************************************************
# * Copyright by ams OSRAM AG                                                 *
# * All rights are reserved.                                                  *
# *                                                                           *
# * IMPORTANT - PLEASE READ CAREFULLY BEFORE COPYING, INSTALLING OR USING     *
# * THE SOFTWARE.                                                             *
# *                                                                           *
# * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       *
# * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT         *
# * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS         *
# * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
# * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,     *
# * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT          *
# * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES LOSS OF USE,      *
# * DATA, OR PROFITS OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY      *
# * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT       *
# * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE     *
# * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.      *
# *****************************************************************************
"""
ZeroMQ server.
"""
import json
import logging
import os
import sys
import time
from pathlib import Path
from threading import Thread
from typing import Callable, Dict, Optional, Union

import zmq

from .tmf8x0x_zeromq_common import (
    DEFAULT_CMD_SERVER_ADDR,
    DEFAULT_RESULT_SERVER_ADDR,
    BaseAdapterInterface,
    CommandId,
    DeviceInfo,
    ErrorCodes,
    HistogramConfig,
    MeasureCommand,
    PersistenceThresholdConfig,
    RequestMessage,
    ResponseMessage,
    ResultContainer,
)
from .tmf8806_regs import tmf8806FactoryCalibData, tmf8806MeasureCmd

logger = logging.getLogger(__name__)

__all__ = [
    "ZeroMqServer",
]


class CalibrationDataStorage:
    """
    Device calibration data storage

    Args:
        path: Folder where the calibration data are stored or
            can be stored.
    """

    def __init__(self, path: Optional[Union[str, Path]] = None) -> None:
        if path is None:
            if sys.platform == "win32":
                path = Path(os.environ["APPDATA"]) / "ams-OSRAM AG"
            else:
                path =  Path(__file__).parent
        self._path = Path(path, "tmf8806_calibration_storage.json")

        self._calib_data = {}
        self._load_data()

    def _load_data(self):
        """Load calibration data from file storage."""
        if self._path.exists():
            try:
                with open(self._path, mode="r", encoding="utf-8") as f:
                    self._calib_data = json.load(f)
            except Exception as exc:
                logger.error("Loading of calibration data failed: %r", exc)

    def _save_data(self):
        """Save calibration data to file storage."""
        with open(self._path, mode="w", encoding="utf-8") as f:
            json.dump(self._calib_data, f)

    @staticmethod
    def _generate_config_key(config: tmf8806MeasureCmd, serial_no: int) -> str:
        """
        Generate a unique key to identify the configuration.
        The key is build from the calibration related configuration
        parameters.

        Args:
            config: Configuration data.

            serial_no: Sensor serial number.

        Returns:
            Unique identifier.
        """
        key = (
            config.data.algo.distanceEnabled,
            config.data.algo.vcselClkDiv2,
            config.data.algo.distanceMode,
            config.data.data.spadSelect,
            config.data.data.spadDeadTime,
        )
        return f"{serial_no}_{''.join(str(e) for e in key)}"

    def store_data(self,
                   calib_data: tmf8806FactoryCalibData,
                   config: tmf8806MeasureCmd,
                   serial_no: int,
                   ) -> None:
        """
        Store calibration data for the given configuration.

        Args:
            calib_data: Calibration data

            config: Configuration related to the calibration data.

            serial_no: Sensor serial number.
        """
        key = self._generate_config_key(config, serial_no)
        self._calib_data[key] = " ".join(f"{e:02x}" for e in bytes(calib_data))
        self._save_data()
        logger.debug("Store calib data for key %s", key)

    def get_data(self,
                 config: tmf8806MeasureCmd,
                 serial_no: int,
                 ) -> Union[tmf8806FactoryCalibData, None]:
        """
        Get calibration data for the given configuration if available.

        Args:
            config: Configuration for which the calibration data is to be read.

            serial_no: Sensor serial number.

        Returns:
            Calibration data or None if not available.
        """
        key = self._generate_config_key(config, serial_no)
        data = self._calib_data.get(key, None)
        if data is not None:
            buffer = b"".join(int(e, 16).to_bytes(1, "little") for e in data.split(" "))
            data = tmf8806FactoryCalibData.from_buffer_copy(buffer)
        logger.debug("Recall calibration for config %s: %s", key, data)
        return data

    def clear_data(self,
                   config: tmf8806MeasureCmd,
                   serial_no: int,
                   ) -> None:
        """
        Clear the calibration data for the given configuration.

        Args:
            config: Configuration for which the calibration data is to be cleared.

            serial_no: Sensor serial number.
        """
        self._calib_data.pop(self._generate_config_key(config, serial_no), None)
        self._save_data()


class ZeroMqServer:
    """
    ZeroMQ server

    The server provides a command and a data socket. The command socket provide
    some services to control the device. The data socket provide measurement results.

    Args:
        adapter: Adapter object which provides device access.
    """

    def __init__(self, adapter: BaseAdapterInterface) -> None:
        self._context = None
        self._cmd_socket = None
        self._result_socket = None
        self._cmd_thread: Optional[Thread] = None
        self._result_thread: Optional[Thread] = None
        self._meas_running = False
        self._registered_commands: Dict[CommandId, Callable] = {}
        self._adapter = adapter
        self._calib_data = CalibrationDataStorage()
        self._curr_device_info = DeviceInfo()

    def _cmd_server(self):
        try:
            while True:
                try:
                    # Wait for next request from client
                    request = RequestMessage(buffer=self._cmd_socket.recv())

                except zmq.ContextTerminated as exc:
                    raise exc

                except Exception as exc:  # pylint: disable=broad-exception-caught
                    logger.error(exc)
                    self._cmd_socket.send(b"\xff")

                else:
                    logger.debug("Received: %s", request)
                    # Send response
                    response = self._process_request(request)
                    logger.debug("Sending : %s", response)
                    self._cmd_socket.send(response.to_buffer())

        except zmq.ContextTerminated:
            logger.info("Command server contexted terminated.")
            self._cmd_socket.close(linger=1)

        except Exception as exc:  # pylint: disable=broad-exception-caught
            logger.error(exc)

        finally:
            logger.debug("Finished command server.")

    def _process_request(self, request: RequestMessage) -> ResponseMessage:
        """
        _summary_

        Args:
            request (_type_): _description_

        Raises:
            exc: _description_

        Returns:
            _type_: _description_
        """
        try:
            resp = ResponseMessage(error_code=ErrorCodes.NO_ERROR)
            if CommandId.IDENTIFY == request.command_id:
                self._curr_device_info = self._adapter.identify()
                resp = ResponseMessage(payload=self._curr_device_info.to_buffer())

            elif CommandId.START_MEASUREMENT == request.command_id:
                config = tmf8806MeasureCmd.from_buffer_copy(request.payload)
                resp = ResponseMessage(error_code=ErrorCodes.NO_ERROR, payload=b"\x00")
                if config.data.command == MeasureCommand.FACTORY_CALIB:
                    # Run factory calibration
                    cal_data = self._adapter.run_factory_calibration(config=config)
                    self._calib_data.store_data(cal_data, config, self._curr_device_info.sensor_serial)

                else:
                    # Start measurement
                    cal_data = self._calib_data.get_data(config, self._curr_device_info.sensor_serial)
                    if cal_data is None:
                        # Clear sensor calibration data by writing default data.
                        cal_data = tmf8806FactoryCalibData()
                        cal_data.id = 2
                    else:
                        resp.payload = b"\x01"
                    config.data.data.factoryCal = 1
                    self._adapter.start_measurement(config=config, cal_data=cal_data)
                    self._start_result_server()

            elif CommandId.CLEAR_CALIBRATION == request.command_id:
                self._calib_data.clear_data(tmf8806MeasureCmd.from_buffer_copy(request.payload),
                                            self._curr_device_info.sensor_serial)

            elif CommandId.GET_CONFIGURATION == request.command_id:
                cfg = self._adapter.get_configuration()
                resp = ResponseMessage(error_code=ErrorCodes.NO_ERROR,
                                       payload=bytes(cfg))

            elif CommandId.STOP_MEASUREMENTS == request.command_id:
                self._stop_result_server()
                try:
                    self._adapter.stop_measurement()
                except Exception as exc:  # pylint: disable=broad-exception-caught
                    logger.error(exc)

            elif CommandId.SET_HISTOGRAM_CFG == request.command_id:
                self._adapter.set_histogram_config(HistogramConfig(buffer=request.payload))

            elif CommandId.GET_HISTOGRAM_CFG == request.command_id:
                cfg = self._adapter.get_histogram_config()
                resp = ResponseMessage(error_code=ErrorCodes.NO_ERROR,
                                       payload=cfg.to_buffer())

            elif CommandId.SET_PERSISTENCE_AND_THRESHOLD == request.command_id:
                self._adapter.set_persistence_threshold_config(
                    PersistenceThresholdConfig(buffer=request.payload))

            elif CommandId.GET_PERSISTENCE_AND_THRESHOLD == request.command_id:
                cfg = self._adapter.get_persistence_threshold_config()
                resp = ResponseMessage(error_code=ErrorCodes.NO_ERROR,
                                       payload=cfg.to_buffer())

            elif CommandId.SET_DRIVER_CFG == request.command_id:
                resp = ResponseMessage(error_code=ErrorCodes.ERROR)

            elif CommandId.GET_DRIVER_CFG == request.command_id:
                resp = ResponseMessage(error_code=ErrorCodes.ERROR)

            elif CommandId.UPDATE_TARGET_BINARIES == request.command_id:
                self._adapter.update_target_binaries(zip_blob=request.payload)

            else:
                resp = ResponseMessage(error_code=ErrorCodes.UNKNOWN_CMD)

        except Exception as exc:  # pylint: disable=broad-exception-caught
            logger.error(exc)
            resp = ResponseMessage(error_code=ErrorCodes.ERROR)

        return resp

    def _result_server(self):
        send_last_result = time.time()
        try:
            while self._meas_running:
                try:
                    result = self._adapter.get_data(timeout=0.05)
                    if result is None :
                        if time.time() - send_last_result > 1.0:
                            # Send empty Frame after 1s
                            logger.debug("Send empty frame")
                            self._result_socket.send(ResultContainer().to_buffer())
                            send_last_result = time.time()
                    else:
                        self._result_socket.send(ResultContainer(data=result).to_buffer())
                        send_last_result = time.time()

                except zmq.ContextTerminated as exc:
                    raise exc

                except Exception as exc:  # pylint: disable=broad-exception-caught
                    logger.error(exc)
                    self._cmd_socket.send(b"Error")

        except zmq.ContextTerminated:
            logger.info("Server contexted terminated.")

        except Exception as exc:  # pylint: disable=broad-exception-caught
            logger.error(exc)

    def _start_result_server(self) -> None:
        """Start the result server."""
        if self._result_thread is None:
            self._meas_running = True
            self._result_thread = Thread(target=self._result_server, name="ZeroMQ result server")
            self._result_thread.start()
            logger.info("Result server started.")

    def _stop_result_server(self) -> None:
        """Stop the result server."""
        self._meas_running = False
        if self._result_thread is not None:
            self._result_thread.join(timeout=5.0)
            if self._result_thread.is_alive():
                logger.error("Result thread is still alive.")
            self._result_thread = None
            logger.info("Result Server stopped")

    def start(self,
              cmd_addr: str = DEFAULT_CMD_SERVER_ADDR,
              result_addr: str = DEFAULT_RESULT_SERVER_ADDR) -> None:
        """
        Start the server

        Args:
            cmd_addr: Address for the command socket.
        """
        self._adapter.open()
        self._curr_device_info = self._adapter.identify()
        if self._cmd_thread is None:
            self._context = zmq.Context()
            self._cmd_socket = self._context.socket(zmq.REP)
            self._result_socket = self._context.socket(zmq.PUB)
            self._cmd_socket.bind(cmd_addr)
            self._result_socket.bind(result_addr)
            self._cmd_thread = Thread(target=self._cmd_server, name="ZeroMQ cmd server")
            self._cmd_thread.start()
            logger.info("Command server started.")

    def stop(self) -> None:
        """
        Stop the server.
        """
        try:
            logger.info("Stop server")
            self._stop_result_server()
            if self._result_socket is not None:
                logger.debug("Closing result socket.")
                self._result_socket.close(linger=1)
                self._result_socket = None
            if self._context is not None:
                logger.debug("Terminating context.")
                self._context.term()
                self._context = None
            if self._cmd_thread is not None:
                logger.info("Wait for command server stop")
                self._cmd_thread.join(timeout=10.0)
                if self._cmd_thread.is_alive():
                    logger.error("Command thread is still alive.")
                self._cmd_thread = None
                self._cmd_socket = None
            logger.debug("Closing adapter")
            self._adapter.close()
            logger.info("Server stopped")
        except Exception as exc:  # pylint: disable=broad-exception-caught
            logger.error(exc)
